--======================================================================
--  ALIENS: A silly little game demonstrating the SDL and mixer libraries
--  Copyright (C) 1998  Sam Lantinga
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program; if not, write to the Free Software
--  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
--  02110-1301 USA.
--
--  Sam Lantinga
--  5635-34 Springhouse Dr.
--  Pleasanton, CA 94588 (USA)
--  slouken@devolution.com
--======================================================================
-- NOTES:
-- * Ported to ysdl by Sean Bolton <musound at jps dot net> 2009
-- * Ported to LuaSDL 0.3.0 by Kein-Hong Man <khman@users.sf.net> 2007
--   See README for more information.
-- * If anything fails, please look for stdout.txt and stderr.txt.
--======================================================================

--======================================================================
-- settings
--======================================================================

local FULLSCREEN = false--true

--======================================================================
-- constants
--======================================================================

local FRAMES_PER_SEC = 50
local PLAYER_SPEED   = 4
local MAX_SHOTS	     = 3
local SHOT_SPEED     = 6
local MAX_ALIENS     = 30
local ALIEN_SPEED    = 5
local ALIEN_ODDS     = 1 * FRAMES_PER_SEC
local EXPLODE_TIME   = 4

--======================================================================
-- functions for closing, error reporting, finding data
--======================================================================

local function stderr(msg) io.stderr:write(msg) end

local function exit() SDL.Quit() end

local function DATAFILE(X)
  -- was: return SDL.LuaSDL_DirApp.."data"..SDL.LuaSDL_DirSep..X
  return "./aliens-data/" .. X
end

--======================================================================
-- global data
--======================================================================

local screen, background
local screen_w, screen_h, background_w, background_h

-- object: alive, facing, x, y, image
local player = {}
local reloading

local shots = {}
for i = 1, MAX_SHOTS do shots[i] = {} end

local aliens = {}
for i = 1, MAX_ALIENS do aliens[i] = {} end

local explosions = {}
for i = 1, MAX_ALIENS+1 do explosions[i] = {} end

local MAX_UPDATES = 3 * (1 + MAX_SHOTS + MAX_ALIENS)

local numupdates

local srcupdate = {}
for i = 1, MAX_UPDATES do
  srcupdate[i] = SDL.newRect()
end
local dstupdate = {}
for i = 1, MAX_UPDATES do
  dstupdate[i] = SDL.newRect()
end

-- blits: src, srcrect, dstrect
local blits = {}
for i = 1, MAX_UPDATES do blits[i] = {} end

local music

local MUSIC_WAV = 0
local SHOT_WAV = 1
local EXPLODE_WAV = 2
local NUM_WAVES = 3

local sounds = {}

--======================================================================
-- resource management
--======================================================================

local function LoadImage(datafile, transparent)
  local image = IMG.Load(datafile)
  if not image then
    stderr("Couldn't load image "..datafile..": "..IMG.GetError().."\n")
    return
  end
  -- if transparent then
  --   -- Assuming 8-bit BMP image
  --   image:SetColorKey(SDL.SRCCOLORKEY + SDL.RLEACCEL, -- should be bitwise-or
  --                     image:getPixel(0, 0))
  -- end
  local surface = image:DisplayFormat()
  image:FreeSurface()
  -- ??? SetColorKey apparently needs to be done on surface, not image
  if transparent then
    surface:SetColorKey(SDL.SRCCOLORKEY + SDL.RLEACCEL, -- should be bitwise-or
                        surface:getPixel(0, 0))
  end
  return surface
end

local function LoadData()
  -- Load sounds
  music = Mix.LoadMUS(DATAFILE("music.it"))
  if not music then
    stderr("Warning: Couldn't load music: "..Mix.GetError().."\n")
  end

  sounds[SHOT_WAV] = Mix.LoadWAV(DATAFILE("shot.wav"))
  sounds[EXPLODE_WAV] = Mix.LoadWAV(DATAFILE("explode.wav"))

  -- Load graphics
  player.image = LoadImage(DATAFILE("player.gif"), true)
  if not player.image then
    return false
  end
  player.image_w, player.image_h = player.image:getSize()
  shots[1].image = LoadImage(DATAFILE("shot.gif"))
  if not shots[1].image then
    return false
  end
  shots[1].image_w, shots[1].image_h = shots[1].image:getSize()
  for i = 2, MAX_SHOTS do
    shots[i].image = shots[1].image
    shots[i].image_w = shots[1].image_w
    shots[i].image_h = shots[1].image_h
  end
  aliens[1].image = LoadImage(DATAFILE("alien.gif"), true)
  if not aliens[1].image then
    return false
  end
  aliens[1].image_w, aliens[1].image_h = aliens[1].image:getSize()
  for i = 2, MAX_ALIENS do
    aliens[i].image = aliens[1].image
    aliens[i].image_w = aliens[1].image_w
    aliens[i].image_h = aliens[1].image_h
  end
  explosions[1].image = LoadImage(DATAFILE("explosion.gif"), true)
  explosions[1].image_w, explosions[1].image_h = explosions[1].image:getSize()
  for i = 2, MAX_ALIENS+1 do
    explosions[i].image = explosions[1].image
    explosions[i].image_w = explosions[1].image_w
    explosions[i].image_h = explosions[1].image_h
  end
  background = LoadImage(DATAFILE("background.gif"))
  background_w, background_h = background:getSize()

  -- Set up the update rectangle pointers
  for i = 1, MAX_UPDATES do
    blits[i].srcrect = srcupdate[i]
    blits[i].dstrect = dstupdate[i]
  end
  return true
end

local function FreeData()
  -- Free sounds
  if music then
    music:FreeMusic()
  end
  for i = 1, NUM_WAVES - 1 do
    sounds[i]:FreeChunk()
  end

  -- Free graphics
  player.image:FreeSurface()
  shots[1].image:FreeSurface()
  aliens[1].image:FreeSurface()
  explosions[1].image:FreeSurface()
  background:FreeSurface()
end

--======================================================================
-- alien spawning function
--======================================================================

local function CreateAlien()
  -- Look for a free alien slot
  local i
  for I = 1, MAX_ALIENS do
    if not aliens[I].alive then
      i = I; break
    end
  end
  if not i then return end

  -- Figure out which direction it travels
  repeat
    aliens[i].facing = math.random(3) - 2
  until aliens[i].facing ~= 0

  -- Figure out it's initial location
  aliens[i].y = 0
  if aliens[i].facing < 0 then
    aliens[i].x = screen_w - aliens[i].image_w - 1
  else
    aliens[i].x = 0
  end
  aliens[i].alive = true
end

--======================================================================
-- screen updating
--======================================================================

local function DrawObject(sprite)
  local update = blits[numupdates]
  numupdates = numupdates + 1
  update.src = sprite.image
  local iw, ih = sprite.image:getSize()
  SDL.setRect(update.srcrect, 0, 0, iw, ih)
  SDL.setRect(update.dstrect, sprite.x, sprite.y, iw, ih)
end

local function EraseObject(sprite)
  -- The background wraps horizontally across the screen
  local update = blits[numupdates]
  numupdates = numupdates + 1
  update.src = background
  local _x = sprite.x % background_w
  local _w = sprite.image_w
  local wrap = (_x + _w) - (background_w)
  if wrap > 0 then
    _w = _w - wrap
  end
  SDL.setRect(update.srcrect, _x, sprite.y, _w, sprite.image_h)
  SDL.setRect(update.dstrect, sprite.x, sprite.y, _w, sprite.image_h)

  -- Assuming sprites can only wrap across one background tile
  if wrap > 0 then
    update = blits[numupdates]
    numupdates = numupdates + 1
    update.src = background
    SDL.setRect(update.srcrect, 0, sprite.y, wrap, sprite.image_h)
    SDL.setRect(update.dstrect, (math.floor(sprite.x / background_w) + 1) * background_w,
                                   sprite.y, wrap, sprite.image_h)
  end
end

local function UpdateScreen()
  for i = 1, numupdates-1 do
    SDL.LowerBlit(blits[i].src, blits[i].srcrect, screen, blits[i].dstrect)
  end
  -- -FIX- ysdl doesn't (yet) have an SDL_UpdateRects, so we just update the whole screen
  -- SDL.SDL_UpdateRects(screen, numupdates-1, dstupdate)
  screen:UpdateRect(0, 0, 0, 0)
  numupdates = 1
end

--======================================================================
-- support functions for game
--======================================================================

local function Collide(sprite1, sprite2)
  if sprite1.y >= (sprite2.y + sprite2.image_h) or
     sprite1.x >= (sprite2.x + sprite2.image_w) or
     sprite2.y >= (sprite1.y + sprite1.image_h) or
     sprite2.x >= (sprite1.x + sprite1.image_w) then
    return false
  end
  return true
end

local next_tick = 0
local function WaitFrame()
  -- Wait for the next frame
  local this_tick = SDL.GetTicks()
  if this_tick < next_tick then
    SDL.Delay(next_tick - this_tick)
  end
  next_tick = this_tick + (1000 / FRAMES_PER_SEC)
end

--======================================================================
-- main game function
--======================================================================

-- This of course can be optimized :-)
local function RunGame()
  local event

  -- Paint the background
  numupdates = 1
  local dst = SDL.newRect()
  local i = 0
  while i < screen_w do
    iw, ih = background:getSize()
    SDL.setRect(dst, i, 0, iw, ih)
    SDL.BlitSurface(background, nil, screen, dst)
    i = i + iw
  end
  screen:UpdateRect(0, 0, 0, 0)

  -- Initialize the objects
  player.alive = true
  player.x = (screen_w - player.image_w) / 2
  player.y = (screen_h - player.image_h) - 1
  player.facing = 0
  DrawObject(player)

  for i = 1, MAX_SHOTS do
    shots[i].alive = false
  end
  for i = 1, MAX_ALIENS do
    aliens[i].alive = false
  end
  CreateAlien()
  DrawObject(aliens[1])
  UpdateScreen()

  while player.alive do
    -- Wait for the next frame
    WaitFrame()

    -- Poll input queue, run keyboard loop
    event = SDL.PollEvent()
    while event do
      if event.type == SDL.QUIT then
        return
      end
      event = SDL.PollEvent()
    end

    -- Erase everything from the screen
    for i = 1, MAX_SHOTS do
      if shots[i].alive then EraseObject(shots[i]) end
    end
    for i = 1, MAX_ALIENS do
      if aliens[i].alive then EraseObject(aliens[i]) end
    end
    EraseObject(player)
    for i = 1, MAX_ALIENS+1 do
      if explosions[i].alive then EraseObject(explosions[i]) end
    end

    -- Decrement the lifetime of the explosions
    for i = 1, MAX_ALIENS+1 do
      if explosions[i].alive then
        explosions[i].alive = explosions[i].alive - 1
        if explosions[i].alive == 0 then
          explosions[i].alive = nil
        end
      end
    end

    -- Create new aliens
    if math.random(ALIEN_ODDS) == 1 then
      CreateAlien()
    end

    -- Create new shots
    if not reloading then
      if SDL.GetKeyState(SDL.K_SPACE) then
        local i
        for I = 1, MAX_SHOTS do
          if not shots[I].alive then
            i = I; break
          end
        end
        if i then
          shots[i].x = player.x + (player.image_w - shots[i].image_w) / 2
          shots[i].y = player.y - shots[i].image_h
          shots[i].alive = true
          Mix.PlayChannel(SHOT_WAV, sounds[SHOT_WAV], 0)
        end
      end
    end
    reloading = SDL.GetKeyState(SDL.K_SPACE)

    -- Move the player
    player.facing = 0
    if SDL.GetKeyState(SDL.K_RIGHT) then
      player.facing = player.facing + 1
    end
    if SDL.GetKeyState(SDL.K_LEFT) then
      player.facing = player.facing - 1
    end
    player.x = player.x + player.facing * PLAYER_SPEED
    if player.x < 0 then
      player.x = 0
    elseif player.x >= (screen_w - player.image_w) then
      player.x = (screen_w - player.image_w) - 1
    end

    -- Move the aliens
    for i = 1, MAX_ALIENS do
      if aliens[i].alive then
        aliens[i].x = aliens[i].x + aliens[i].facing * ALIEN_SPEED
        if aliens[i].x < 0 then
          aliens[i].x = 0
          aliens[i].y = aliens[i].y + aliens[i].image_h
          aliens[i].facing = 1
        elseif aliens[i].x >= (screen_w - aliens[i].image_w) then
          aliens[i].x = (screen_w - aliens[i].image_w) - 1
          aliens[i].y = aliens[i].y + aliens[i].image_h
          aliens[i].facing = -1
        end
      end
    end

    -- Move the shots
    for i = 1, MAX_SHOTS do
      if shots[i].alive then
        shots[i].y = shots[i].y - SHOT_SPEED
        if shots[i].y < 0 then
          shots[i].alive = false
        end
      end
    end

    -- Detect collisions
    for j = 1, MAX_SHOTS do
      for i = 1, MAX_ALIENS do
        if shots[j].alive and aliens[i].alive and Collide(shots[j], aliens[i]) then
          aliens[i].alive = false
          explosions[i].x = aliens[i].x
          explosions[i].y = aliens[i].y
          explosions[i].alive = EXPLODE_TIME
          Mix.PlayChannel(EXPLODE_WAV, sounds[EXPLODE_WAV], 0)
          shots[j].alive = false
          break
        end
      end
    end
    for i = 1, MAX_ALIENS do
      if aliens[i].alive and Collide(player, aliens[i]) then
        aliens[i].alive = false
        explosions[i].x = aliens[i].x
        explosions[i].y = aliens[i].y
        explosions[i].alive = EXPLODE_TIME
        player.alive = false
        explosions[MAX_ALIENS].x = player.x
        explosions[MAX_ALIENS].y = player.y
        explosions[MAX_ALIENS].alive = EXPLODE_TIME
        Mix.PlayChannel(EXPLODE_WAV, sounds[EXPLODE_WAV], 0)
      end
    end

    -- Draw the aliens, shots, player, and explosions
    for i = 1, MAX_ALIENS do
      if aliens[i].alive then DrawObject(aliens[i]) end
    end
    for i = 1, MAX_SHOTS do
      if shots[i].alive then DrawObject(shots[i]) end
    end
    if player.alive then DrawObject(player) end
    for i = 1, MAX_ALIENS+1 do
      if explosions[i].alive then DrawObject(explosions[i]) end
    end
    UpdateScreen()

    -- Loop the music
    if music and not Mix.PlayingMusic() then
      Mix.PlayMusic(music, 0)
    end

    -- Check for keyboard abort
    if SDL.GetKeyState(SDL.K_ESCAPE) then
      player.alive = false
    end
  end--while

  -- Wait for the player to finish exploding
  while Mix.Playing(EXPLODE_WAV) do
    WaitFrame()
  end

  Mix.HaltChannel(-1)
end

--======================================================================
-- main program starts here
--======================================================================

-- Initialize the SDL library
if SDL.Init(SDL.INIT_AUDIO + SDL.INIT_VIDEO) < 0 then -- should be bitwise-or
  stderr("Couldn't initialize SDL: "..SDL.GetError().."\n")
  return
end

-- Open the audio device
if Mix.OpenAudio(11025, SDL.AUDIO_U8, 1, 512) < 0 then
  stderr("Warning: Couldn't set 11025 Hz 8-bit audio\n- Reason: "..Mix.GetError().."\n")
  exit()
end

-- Open the display device
local flags = SDL.SWSURFACE
if FULLSCREEN then
  flags = flags + SDL.FULLSCREEN -- should be bitwise-or
end

screen = SDL.SetVideoMode(640, 480, 0, flags)
if not screen then
  stderr("Couldn't set 640x480 video mode: "..SDL.GetError().."\n")
  exit()
end
screen_w, screen_h = screen:getSize()

-- Initialize the random number generator
math.randomseed(os.time())

-- Load the music and artwork
if LoadData() then
  -- Run the game
  RunGame()
  -- Free the music and artwork
  FreeData()
end

-- Quit
Mix.CloseAudio()
exit()

